import re
from datetime import timedelta

EXPIRE_FORMAT = "%Y-%m-%d %H:%M:%S"
DELTA_REGEX = re.compile(
    r"^((?P<weeks>[\.\d]+?)w)? *"
    r"^((?P<days>[\.\d]+?)d)? *"
    r"((?P<hours>[\.\d]+?)h)? *"
    r"((?P<minutes>[\.\d]+?)m)? *"
    r"((?P<seconds>[\.\d]+?)s?)?$"
)


# copied from https://github.com/zeitgitter/zeitgitterd/blob/master/zeitgitter/deltat.py
def parse_delta(delta_str):
    """
    Parse a time string e.g. '2h 13m' or '1.5d' into a timedelta object.
    Based on Peter's answer at https://stackoverflow.com/a/51916936/2445204
    and virhilo's answer at https://stackoverflow.com/a/4628148/851699
    :param delta_str: A string identifying a duration, e.g. '2h13.5m'
    :return datetime.timedelta: A datetime.timedelta object
    """
    parts = DELTA_REGEX.match(delta_str)
    assert (
        parts is not None
    ), """Could not parse any time information from '{}'.
    Examples of valid strings: '8h', '2d 8h 5m 2s', '2m4.3s'""".format(
        delta_str
    )
    time_params = {
        name: float(param) for name, param in parts.groupdict().items() if param
    }
    return timedelta(**time_params)



import platform
import sqlite3
from pathlib import Path

from grantme.constants.misc import PROG
from grantme.constants.sqls import CREATE_DB, CREATE_INDEX
from grantme.utils.printers import info


UPDATE_MOTD_DIR = Path("/etc/update-motd.d")


def create_motd():
    if "ubuntu" not in platform.version().lower():
        info(
            "Not running Ubuntu distributions, skipping creating the update-motd"
            " fragment."
        )
        return

    existing_frag_nums = {
        path.name[:2] for path in Path("/etc/update-motd.d").iterdir()
    }

    for i in range(99, -1, -1):
        i_str = f"{i:02}"
        if i_str in existing_frag_nums:
            continue

        motd_frag_name = f"{i_str}-grantme-status"
        motd_frag_file = UPDATE_MOTD_DIR / motd_frag_name
        with open(str(motd_frag_file), "w") as f:
            f.writelines(["#!/bin/sh\n", "echo\n", f"{PROG} --status\n"])

        # make new MOTD fragment executable
        motd_frag_file.chmod(motd_frag_file.stat().st_mode | 0o111)

        break


def create_db(data_dir: Path, db_file: Path):
    data_dir.mkdir(parents=True, exist_ok=True)
    conn = sqlite3.connect(str(db_file))
    setup_db(conn)
    conn.close()

    # make db accessible to everyone
    data_dir.chmod(data_dir.stat().st_mode | 0o666)
    db_file.chmod(db_file.stat().st_mode | 0o666)


def setup_db(conn: sqlite3.Connection):
    cur = conn.cursor()
    cur.execute(CREATE_DB)
    cur.execute(CREATE_INDEX)
    cur.close()
    conn.commit()



#!/usr/bin/env python3

import sqlite3
from dataclasses import dataclass
from datetime import datetime
from enum import Enum, auto
from typing import Mapping, Optional

from grantme.constants.msgs import *
from grantme.constants.sqls import *
from grantme.utils.errors import *

from .utils import printers, timer


class Mode(Enum):
    EXCLUSIVE = auto()
    SHARED = auto()


@dataclass
class LockStatus:
    mode: Mode
    user_expires: Mapping[str, datetime]


@dataclass
class GrantResult:
    curr_mode: Optional[Mode]
    prev_mode: Optional[Mode]
    curr_expire_at: Optional[datetime]
    prev_expire_at: Optional[datetime]


def write_lock(who: str, mode: Mode, expire_at: datetime, conn: sqlite3.Connection):
    cur = conn.cursor()
    cur.execute(
        UPSERT,
        {"who": who, "shared": mode == Mode.SHARED, "expire_at": expire_at},
    )
    assert cur.rowcount == 1
    cur.close()


def get_status(conn: sqlite3.Connection) -> Optional[LockStatus]:
    now = datetime.now()

    cur = conn.cursor()
    cur.execute(QUERY_SHARED, {"now": now})
    rows = cur.fetchall()
    if len(rows) > 0:
        cur.close()
        return LockStatus(
            mode=Mode.SHARED,
            user_expires={row["user"]: row["expire_at"] for row in rows},
        )

    cur.execute(QUERY_EXCLUSIVE, {"now": now})
    row = cur.fetchone()
    if row is not None:
        cur.close()
        return LockStatus(
            mode=Mode.EXCLUSIVE, user_expires={row["user"]: row["expire_at"]}
        )

    cur.close()
    return None


def revoke(who, conn: sqlite3.Connection) -> Optional[Mode]:
    status = get_status(conn)

    if status is None:
        return None

    if who not in status.user_expires:
        return None

    cur = conn.cursor()
    cur.execute(RELEASE, {"who": who})
    assert cur.rowcount == 1
    cur.close()
    conn.commit()

    return status.mode


def grant(
    who: str, wanted: Mode, duration: str, conn: sqlite3.Connection
) -> GrantResult:
    attempted_expire = datetime.now() + timer.parse_delta(duration)

    status = get_status(conn)
    result = GrantResult(None, None, None, None)

    # Case 1: if no lock is granted, simply grant one
    if status is None:
        write_lock(who, wanted, attempted_expire, conn)
        conn.commit()

        result.curr_mode = wanted
        result.curr_expire_at = attempted_expire
        return result

    # Case 2: if neither s- nor x-locked, try lock
    if who not in status.user_expires:
        # cannot want shared when an exclusive is granted (not to `who`)
        # cannot want exclusive when any other lock is granted (not to `who`)
        if (wanted == Mode.SHARED and status.mode == Mode.EXCLUSIVE) or (
            wanted == Mode.EXCLUSIVE and len(status.user_expires) > 0
        ):
            raise GrantmeError(GRANT_FAILURE_MSG, GRANT_FAILURE_CODE)

        write_lock(who, wanted, attempted_expire, conn)
        conn.commit()

        result.curr_mode = wanted
        result.curr_expire_at = attempted_expire
        return result

    # Case 3: if already locked, try (upgrade and) prolong
    result.prev_mode = status.mode

    # prevent accidental downgrade
    if result.prev_mode == Mode.EXCLUSIVE and wanted == Mode.SHARED:
        wanted = Mode.EXCLUSIVE

    # try upgrade
    if result.prev_mode == Mode.SHARED and wanted == Mode.EXCLUSIVE:
        if len(status.user_expires) > 1:
            raise GrantmeError(UPGRADE_FAILURE_MSG, UPGRADE_FAILURE_CODE)

        cur = conn.cursor()
        cur.execute(UPGRADE, {"who": who})
        assert cur.rowcount == 1
        cur.close()

        result.curr_mode = Mode.EXCLUSIVE
    else:
        result.curr_mode = result.prev_mode

    # try prolong
    old_expire = status.user_expires[who]
    result.prev_expire_at = old_expire
    if old_expire > attempted_expire:
        result.curr_expire_at = old_expire
    else:
        write_lock(who, wanted, attempted_expire, conn)
        result.curr_expire_at = attempted_expire

    conn.commit()
    return result


def print_status(you: str, conn: sqlite3.Connection):
    status = get_status(conn)

    printers.show_sep_line()

    if status is None:
        printers.show_no_lock()
    elif status.mode == Mode.EXCLUSIVE:
        user, expire_at = next(iter(status.user_expires.items()))
        if user == you:
            printers.show_granted(
                "Exclusive", expire_at, highlighted_user=user + " (YOU)"
            )
        else:
            printers.show_granted("Exclusive", expire_at, others=[user])
            printers.warning(DO_NOT_DISTURB)
    else:
        last_expire = max(status.user_expires.values())
        if you in status.user_expires:
            others = list(status.user_expires.keys())
            others.remove(you)
            printers.show_granted(
                "Shared", last_expire, highlighted_user=you + " (YOU)", others=others
            )
            printers.show_expire(status.user_expires[you], "Your")
        else:
            printers.show_granted(
                "Shared", last_expire, others=status.user_expires.keys()
            )
            printers.slock_guide()

    printers.show_sep_line()


def revoke_and_print_result(who, conn: sqlite3.Connection):
    result = revoke(who, conn)
    if result is None:
        printers.warning(NOTHING_REVOKED)
    elif result == Mode.EXCLUSIVE:
        printers.show_revoke_ok("exclusive", who)
    else:
        printers.show_revoke_ok("shared", who)


def grant_and_print_result(
    who: str, want_shared: bool, duration: str, conn: sqlite3.Connection
):
    result = grant(who, Mode.SHARED if want_shared else Mode.EXCLUSIVE, duration, conn)

    # when called with a lock already held
    if result.prev_mode is not None:
        assert result.curr_mode is not None
        assert result.prev_expire_at is not None
        assert result.curr_expire_at is not None

        if result.prev_mode == Mode.EXCLUSIVE and want_shared:
            assert result.curr_mode == Mode.EXCLUSIVE
            printers.warning(ATTEMPT_SLOCK_WITH_XLOCK)

        if result.prev_mode == Mode.SHARED and result.curr_mode == Mode.EXCLUSIVE:
            printers.ok(UPGRADE_OK)

        if result.prev_expire_at == result.curr_expire_at:
            printers.warning(ATTEMPT_SHORTER_DURATION_THAN_ALREADY_GRANTED)
        else:
            printers.ok(PROLONG_OK)
        printers.show_expire(result.curr_expire_at, "Your")

    # when called without a lock already held
    else:
        if result.curr_mode is not None:
            assert result.prev_expire_at is None
            assert result.curr_expire_at is not None
            printers.show_grant_ok("shared" if want_shared else "exclusive", who)
            printers.show_expire(result.curr_expire_at, "Your")



# Help messages
from grantme.constants.misc import PROG


PROG_HELP = (
    'This is a utility for "granting" exclusive/shared access to '
    "machines. It is purely advisory in the sense that neither permissions "
    "is necessarily required to access the host. Its purpose is to make you "
    "aware of the presence of other active users so that you don't disrupt "
    "their work.\n\n"
    "Use `{prog} --mode {{exclusive,shared}} --duration 2h30m to obtain/prolong"
    "permission for 2 hours and 30 minutes.\n"
    "Use `{prog} --revoke` to invalidate your granted permission.\n"
    "Use `{prog} --revoke --user USER` to invalidate other users' granted"
    "permission.\n"
    "Use `{prog} --status` to view permissions of yours and other users'.".format(
        prog=PROG
    )
)
DURATION_HELP = (
    'Expected access duration, formatted as "?w?d?h?m?s". Default to "30m". You can'
    " rerun {prog} to prolong permission.".format(prog=PROG)
)
USER_HELP = (
    "The user whose permissions should be revoked. Defaults to the current login user."
)
INIT_HELP = (
    "Initialize the backend database. Root privilege might be needed (i.e., `sudo`)."
)

# Info
SLOCK_CMD = "{prog} --mode shared --duration ?d?h?m?s".format(prog=PROG)
XLOCK_CMD = "{prog} --mode exclusive --duration ?d?h?m?s".format(prog=PROG)
PROLONG_OK = "Permission is successfully prolonged."
UPGRADE_OK = "Shared permission upgraded to an exclusive permission."

# Warnings
DO_NOT_DISTURB = "Please **AVOID DEMANDING TASKS**, so as to not disrupt others."
NOTHING_REVOKED = "No permission is currently granted; nothing is revoked."
ATTEMPT_SLOCK_WITH_XLOCK = (
    "An exclusive permission already exists. Ignoring `--mode shared`."
)
ATTEMPT_SHORTER_DURATION_THAN_ALREADY_GRANTED = (
    "A permission that expires later than the given duration already exists."
    "No change is made."
)



CREATE_DB = (
    "create table if not exists leases ("
    "  user      TEXT constraint leases_pk primary key,"
    "  shared    INTEGER not null,"
    "  expire_at timestamp not null"
    ");"
)
CREATE_INDEX = (
    "create index if not exists leases_expire_at_index on leases (expire_at);"
)
QUERY_SHARED = (
    "select user, expire_at from leases where shared = true and expire_at > :now;"
)
QUERY_EXCLUSIVE = (
    "select user, expire_at from leases where shared = false and expire_at > :now;"
)
UPSERT = (
    "insert into leases values (:who, :shared, :expire_at)"
    "  on conflict(user) do update set shared=:shared, expire_at=:expire_at;"
)
UPGRADE = "update leases set shared = false where user = :who and shared = true;"
RELEASE = "delete from leases where user = :who;"



from grantme.constants.msgs import SLOCK_CMD, XLOCK_CMD


# Formatting codes
# copied from https://stackoverflow.com/a/287944
class FormatCode:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    ENDC = "\033[0m"


def show_sep_line():
    print(FormatCode.HEADER, "#" * 80, FormatCode.ENDC, sep="")


def slock_guide():
    print("You can obtain shared access with")
    print("\t" + SLOCK_CMD)


def xlock_guide():
    print("You can obtain exclusive access with")
    print("\t" + XLOCK_CMD)


def show_no_lock():
    print("Neither exclusive nor shared access is not granted yet.")
    slock_guide()
    xlock_guide()


def show_expire(expire, whose=None):
    if whose is None:
        subject = "The permission"
    else:
        subject = "{whose} permission".format(whose=whose)
    print(
        "{subject} expires at ".format(subject=subject)
        + FormatCode.BOLD
        + str(expire)
        + FormatCode.ENDC
        + "."
    )


def show_granted(mode, expire, highlighted_user=None, others=[]):
    print(mode, "access is granted to")
    if highlighted_user is None:
        who = ", ".join(others)
    else:
        who = FormatCode.BOLD + highlighted_user + FormatCode.ENDC
        if len(others) > 0:
            who += ", "
            who += ", ".join(others)
    print("\t" + who + ",")
    print(
        "(the last of) which expires at "
        + FormatCode.BOLD
        + str(expire)
        + FormatCode.ENDC
        + "."
    )


def show_grant_ok(mode, who):
    ok("{who}'s {mode} permission is successfully granted.".format(who=who, mode=mode))


def show_revoke_ok(mode, who):
    ok("{who}'s {mode} permission is successfully revoked.".format(who=who, mode=mode))


def ok(msg):
    print(FormatCode.OKGREEN + msg + FormatCode.ENDC)


def info(msg):
    print(FormatCode.OKBLUE + msg + FormatCode.ENDC)


def warning(msg):
    print(FormatCode.WARNING + msg + FormatCode.ENDC)


# When some op fails within expectation
def failure(msg):
    print(FormatCode.FAIL + msg + FormatCode.ENDC)


# Unexpected errors
def error(msg):
    print(FormatCode.FAIL + msg + FormatCode.ENDC)



import sqlite3
import unittest


from grantme.procedures import Mode, get_status, grant, revoke
from grantme.utils.setup import setup_db


class SingleUserTest(unittest.TestCase):
    def setUp(self):
        self.conn = sqlite3.connect(":memory:", detect_types=sqlite3.PARSE_DECLTYPES)
        self.conn.row_factory = sqlite3.Row
        setup_db(self.conn)

        self.me = "cook"

    def tearDown(self):
        self.conn.close()

    def assert_emptiness(self):
        status = get_status(self.conn)
        self.assertIsNone(status)

    def assert_the_only(self, who, mode):
        status = get_status(self.conn)
        self.assertIsNotNone(status)
        self.assertEqual(status.mode, mode)
        self.assertEqual(len(status.user_expires), 1)
        self.assertIn(who, status.user_expires)

    def test_acquire_and_revoke_shared(self):
        grant(self.me, Mode.SHARED, "1m", self.conn)
        self.assert_the_only(self.me, Mode.SHARED)

        revoke(self.me, self.conn)
        self.assert_emptiness()

    def test_acquire_and_revoke_exclusive(self):
        grant(self.me, Mode.EXCLUSIVE, "1m", self.conn)
        self.assert_the_only(self.me, Mode.EXCLUSIVE)

        revoke(self.me, self.conn)
        self.assert_emptiness()

    def test_upgrade(self):
        grant(self.me, Mode.SHARED, "1m", self.conn)
        self.assert_the_only(self.me, Mode.SHARED)

        grant(self.me, Mode.EXCLUSIVE, "1m", self.conn)
        self.assert_the_only(self.me, Mode.EXCLUSIVE)

        revoke(self.me, self.conn)
        self.assert_emptiness()

    def test_no_downgrade(self):
        grant(self.me, Mode.EXCLUSIVE, "1m", self.conn)
        self.assert_the_only(self.me, Mode.EXCLUSIVE)

        grant(self.me, Mode.SHARED, "1m", self.conn)
        self.assert_the_only(self.me, Mode.EXCLUSIVE)

        revoke(self.me, self.conn)
        self.assert_emptiness()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import argparse
import sqlite3
import sys
from getpass import getuser
from pathlib import Path
from traceback import print_tb

from grantme.constants.misc import PROG
from grantme.constants.msgs import DURATION_HELP, INIT_HELP, PROG_HELP, USER_HELP
from grantme.procedures import (
    grant_and_print_result,
    print_status,
    revoke_and_print_result,
)
from grantme.utils.errors import UNEXPECTED_ERROR_CODE, GrantmeError
from grantme.utils.printers import error
from grantme.utils.setup import create_db, create_motd


DB_FILENAME = "db.sqlite3"
DATA_DIR = Path("/var/lib/grantme")
DB_FILE = DATA_DIR / DB_FILENAME


def main():
    current_user = getuser()

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(PROG_HELP),
    )
    parser.add_argument("-s", "--status", action="store_true")
    parser.add_argument("-m", "--mode", choices=["exclusive", "shared"])
    parser.add_argument("-d", "--duration", default="30m", help=DURATION_HELP)
    parser.add_argument("-r", "--revoke", action="store_true")
    parser.add_argument("-u", "--user", default=current_user, help=USER_HELP)
    parser.add_argument("-i", "--init", action="store_true", help=INIT_HELP)

    args = parser.parse_args()

    if args.init:
        try:
            create_db(DATA_DIR, DB_FILE)
        except Exception as err:
            error(
                "Failed to initiate the database. Please ensure that you have root"
                f" privilege for creating {DB_FILE} file. To resolve corrupted"
                f" database files, you may remove {DB_FILE} and then run"
                f" `{PROG} --init` to create a new one."
            )
            error(f"Original exception: {err}")
            sys.exit(UNEXPECTED_ERROR_CODE)

        create_motd()

    try:
        exit_code = 0

        conn = sqlite3.connect(str(DB_FILE), detect_types=sqlite3.PARSE_DECLTYPES)
        conn.row_factory = sqlite3.Row

        if args.init:
            pass  # to prevent printing CLI help message... should fix parser later!
        elif args.status:
            print_status(current_user, conn)
        elif args.revoke:
            revoke_and_print_result(args.user, conn)
        elif args.mode is not None:
            assert args.mode in ("shared", "exclusive")
            grant_and_print_result(
                current_user, args.mode == "shared", args.duration, conn
            )
        else:
            parser.print_help()
    except GrantmeError as err:
        error(err.msg)
        exit_code = err.code
    except Exception as err:
        error(f"Unexpected exception: {err}")
        print_tb(err.__traceback__)
        exit_code = UNEXPECTED_ERROR_CODE
    finally:
        conn.close()
        sys.exit(exit_code)


if __name__ == "__main__":
    main()



class GrantmeError(RuntimeError):
    def __init__(self, msg, code):
        super().__init__(msg)

        self.msg = msg
        self.code = code


# Error messages and codes
UNEXPECTED_ERROR_CODE = 255

UPGRADE_FAILURE_MSG = "Failed to upgrade shared permission to exclusive permission."
UPGRADE_FAILURE_CODE = 6
GRANT_FAILURE_MSG = "Failed to grant permission."
GRANT_FAILURE_CODE = 7
